/* z_io.c (parallel io)

   Ambros Marzetta

 * $Source: /home/schilte/marzetta/work/distr/zram/RCS/z_io.c,v $
 * $Revision: 1.4 $
 * $Date: 1996/06/03 19:00:48 $
 * $Log: z_io.c,v $
 * Revision 1.4  1996/06/03  19:00:48  marzetta
 * New definition of z_NewNode.
 *
 * Revision 1.3  1996/05/29  17:30:22  marzetta
 * Removed some unused variables in the sequential version
 *
 * Revision 1.2  1996/05/15  16:54:49  marzetta
 * \Exported z_io_AppendProc
 *
 * Revision 1.1  1996/05/08  20:29:30  marzetta
 * Initial revision
 *
 */

#include "z_io.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "z_par.h"
#include "errno.h"
#include "z_md.h"


typedef struct z_FamilyNode {
  Z_NODE_HEADER;
  z_NodeP ch[10];
} *z_FamilyNodeP;

#ifndef SEQUENTIAL
static z_JobNum RedirectStdoutJobNum;
#endif /* SEQUENTIAL */
static boolean io_inited;
static long headersize;		/* size of z_Node header without pointers */


/* Open / Close -------------------------------------------------------- */

void z_io_AppendProc(char *s, const char *path, int numprocs, int myproc)
/* Append the processor number to the path (s := path.nnn). */
{
  /* Append the processor number to the path. */
  if (numprocs == 1)
    sprintf(s, "%s", path);
  else if (numprocs < 10)
    sprintf(s, "%s.%d", path, myproc);
  else if (numprocs < 100)
    sprintf(s, "%s.%02d", path, myproc);
  else
    sprintf(s, "%s.%03d", path, myproc);
} /* z_io_AppendProc */

FILE *z_fopen(const char *path, const char *mode)
{
  /* Open the file named by the path parameter with a processor number
     dependent extension */
  char s[PATH_MAX];

  z_io_AppendProc(s, path, z_numprocs, z_myproc);
  return fopen(s, mode);
} /* *z_fopen */

int z_io_MakeBackup(const char *path)
{
  /* Rename the file path.nnn to path.nnn.bak where nnn is the processor
     number. */
  char from[PATH_MAX], to[PATH_MAX];
  int result;

  z_io_AppendProc(from, path, z_numprocs, z_myproc);
  strcpy(to, from);
  strcat(to, ".bak");
  result = rename(from, to);
#ifndef macintosh
  if (result && errno != ENOENT) perror("rename");
#endif /* macintosh */
  return result;
} /* z_io_MakeBackup */

#ifndef SEQUENTIAL
static void RedirectStdoutJob(void)
{
  char s[PATH_MAX], path[PATH_MAX];
  int len;
  FILE *result;

  len = z_GetMINTBuffer();
  z_GetDataBuffer(path, len);
  path[len] = 0;
  z_io_AppendProc(s, path, z_numprocs, z_myproc);
  result = freopen(s, "w", stdout);
  if (! result) {
    sprintf(path, "z_RedirectStdoutJob(%s)", s);
    perror(path);
    abort();
  }
} /* RedirectStdoutJob */
#endif /* SEQUENTIAL */

void z_io_RedirectStdout(const char *path)
{
  /* Redirect stdout to a file (different for each processor) */
#ifndef SEQUENTIAL
  int len;
  char s[PATH_MAX];
#endif /* SEQUENTIAL */
  
  assert(io_inited);
  assert(z_frontend);
#ifdef SEQUENTIAL
  freopen(path, "w", stdout);
#else /* SEQUENTIAL */
  len = strlen(path);
  assert(len <= PATH_MAX-4);
  z_AddMINTBuffer(len);
  z_AddDataBuffer(path, len);
  z_StartParallelJob(RedirectStdoutJobNum, NULL);
  if (z_myproc == -1) {
    sprintf(s, "%s.fe", path);
    freopen(s, "w", stdout);
  }
  z_WaitWork();
#endif /* SEQUENTIAL */
} /* z_io_RedirectStdout */


/* Read / Write -------------------------------------------------------- */

#ifdef cenju3
#define MAXREAD 8192
#else
#define MAXREAD 2147483647
#endif

long z_fread(void *pointer, long size, long num_items, FILE *stream)
{
    /* Same as fread() but aborts on error and eof */
    long read, r;
	
    size *= num_items;
    read = 0;
    while (read < size) {
	r = fread((char *)pointer + read, sizeof(char),
		  (size - read > MAXREAD ? MAXREAD : size - read), stream);
	if (!r) {
	    if (feof(stream))
		fprintf(stderr, "[%2d]z_fread: end of file\n", z_myproc);
	    else fprintf(stderr, "[%2d]z_fread: error %d\n",
			 z_myproc, ferror(stream));
	    z_md_abort();
	}
	read += r;
    }
    return num_items;
} /* z_fread */

void z_WriteNode(FILE *f, z_NodeP node)
{
  /* Write the node to the file. node may be NULL. */
  z_FamilyNodeP src;
  char *body;  /* body = node without header and children pointers */
  long bodylen;			/* length of body */
  int i;
  boolean notNull;
  
  assert(io_inited);
  if (node) {
    notNull = 1;
    fwrite(&notNull, sizeof(boolean), 1, f);
    src = (z_FamilyNodeP) node;
    body = (char *) & src->ch[src->children];
    bodylen = ((long) src) + node->len - ((long) body);
    /*printf("write node %d %ld %ld\n", node->tag, headersize, bodylen);*/
    fwrite(node, headersize, 1, f); /* write the header */
    fwrite(body, bodylen, 1, f); /* write the body */
    { /* copy the children */
      for (i=0; i<(int)src->children; i++) z_WriteNode(f, src->ch[i]);
    }
  } else {
    notNull = 0;
    fwrite(&notNull, sizeof(boolean), 1, f);
  }
} /* z_WriteNode */

z_NodeP z_ReadNode(FILE *f)
{
  /* recursive */
  struct z_Node head;
  z_FamilyNodeP np;
  long bodylen;			/* length of body */
  boolean notNull;
  char *body;  /* body = node without header and children pointers */
  unsigned long bytes;

  assert(io_inited);
  assert(sizeof(head) >= headersize);
  z_fread(&notNull, sizeof(boolean), 1, f);
  if (notNull) {
    z_fread(&head, headersize, 1, f);
    np = (z_FamilyNodeP) z_NewNode(head.tag, head.len);
    assert(np);
    (void) memcpy(np, &head, headersize);
    body = (char *) & np->ch[np->children];
    bodylen = ((long) np) + np->len - ((long) body);
    /* if (!z_myproc)
	fprintf(stderr, "[%d-%d %lx, %d, %ld]",
		head.tag, np->tag, &np->ch[np->children],
		np->children, bodylen); */
    bytes = z_fread(&np->ch[np->children], 1, bodylen, f);
    if (bytes < (unsigned)bodylen) {
	printf("+++%ld++\n", bytes);
	fflush(stdout);
	fprintf(stderr, "Error in fread 3 :%d-%d\n", ferror(f), feof(f));
    }
    { /* copy the children */
      int i;
      for (i=0; i<(int)np->children; i++) np->ch[i] = z_ReadNode(f);
    }
    {
      np->parent = NULL;
      np->child = NULL;
      np->sibling = NULL;
    }
    return (z_NodeP) np;
  } else
    return NULL;
} /* z_ReadNode */


/* Initialization ------------------------------------------------------ */
  
void z_io_Init(void)
{
  if (! io_inited) {
    io_inited = 1;
#ifndef SEQUENTIAL
    RedirectStdoutJobNum = z_InstallJob(RedirectStdoutJob);
#endif /* SEQUENTIAL */
    headersize = (long)&((z_NodeP)NULL)->parent;
    if ((long)&((z_NodeP)NULL)->sibling < headersize)
      headersize = (long)&((z_NodeP)NULL)->sibling;
  }
} /* z_io_Init */
